/* Subroutines for insn-output.c for GNU compiler.  OpenRISC 1000 version.
   Copyright (C) 1987, 1992, 1997 Free Software Foundation, Inc
   Contributed by Damjan Lampret <lampret@opencores.org> in 1999.
   Based upon the elxsi (which was based upon the vax port) and sparc port.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include "rtl.h"
#include "tree.h"
#include "flags.h"
#include "regs.h"
#include "hard-reg-set.h"

/*
 * for holding comparison operands.
 */
rtx or16_cmp_op0, or16_cmp_op1;
enum machine_mode or16_cmp_mode;

/*
 * used in function prologue/epilogue generation
 */
extern int leaf_function;

int
print_operand_punct_valid_p(code)
     int code;
{
  switch(code)
    {
      case '(':   /* idea taken from sparc; output nop for %( if
                     not optimizing or the slot is not filled. */
      case '%':
        return 1;
    }
  return 0;
}
                                                  
print_operand_address (file, addr)
     FILE *file;
     register rtx addr;
{
  register rtx reg1, reg2, breg, ireg;
  rtx offset;
 
 retry:
  switch (GET_CODE (addr))
    {

    case MEM:
      if (GET_CODE (XEXP (addr, 0)) == REG)
        fprintf (file, "%s", reg_names[REGNO (addr)]);
      else abort();
      break;

    case REG:
      fprintf (file, "0(%s)", reg_names[REGNO (addr)]);
      break;

    case PLUS:
      reg1 = 0; reg2 = 0;
      ireg = 0; breg = 0;
      offset = 0;
      if (GET_CODE (XEXP (addr, 0)) == REG)
        {
          offset = XEXP (addr, 1);
          addr = XEXP (addr, 0);
        }
      else if (GET_CODE (XEXP (addr, 1)) == REG)
        {
          offset = XEXP (addr, 0);
          addr = XEXP (addr, 1);
        }
      output_address (offset);
      fprintf (file, "(%s)", reg_names[REGNO (addr)]);
      break;

    default:
      /* fprintf(file, "{%d}", GET_CODE (addr)); */
      output_addr_const (file, addr);
    }
}

char *
get_function_name()
{
  char *kind = "function";
  if (current_function_decl != 0
      && TREE_CODE (TREE_TYPE (current_function_decl)) == METHOD_TYPE)
    kind = "method";

  if (current_function_decl == NULL)
    {
      return "top level";
    }
  else
    {
      char *name = (*decl_printable_name) (current_function_decl, &kind);
      return name;
    }
}

/*
 * operands[0] will be the branch target.
 * test will be the comparison type.
 * or16_cmp_mode is SImode
 * or16_cmp_op0 and or16_cmp_op1 are the things to be 
 * compared.
 *
 * This routine has to generate a compare and a branch rtx.
 *
 */
int
gen_cond_branch(operands, test)
     rtx operands[];
     enum rtx_code test;
{
  rtx tmp;

  switch(or16_cmp_mode)
    {
    case SImode:
      /* generate a sfXX instruction. */

      /* check if operand 1 is an integer that is too big;
         move it into a register if it is. */
      if(GET_CODE(or16_cmp_op1) == CONST_INT)
        {
          int c = INTVAL(or16_cmp_op1);
          or16_cmp_op1 = force_reg(SImode, or16_cmp_op1);
        }

      /* emit the set instruction. */
      tmp = gen_rtx(SET, VOIDmode, cc0_rtx,
                         gen_rtx(test, SImode,
                                 or16_cmp_op0, or16_cmp_op1));
      emit_insn( tmp);
  
      /* emit the branch instruction. */
      tmp = gen_rtx(SET, VOIDmode, 
                              pc_rtx,
                              gen_rtx(IF_THEN_ELSE, VOIDmode,
                                      gen_rtx(NE, VOIDmode, const0_rtx,
                                              cc0_rtx),
                                      gen_rtx(LABEL_REF,VOIDmode,operands[0]), 
                                      pc_rtx));
      emit_jump_insn( tmp);
      break;
    }

}

int
sym_ref_mem_operand(op, mode)
  register rtx op;
  enum machine_mode mode;
{
  if(GET_CODE(op) == MEM) {
    rtx t1 = XEXP(op, 0);
    if (GET_CODE(t1) == SYMBOL_REF)
      return 1;
  }

  return 0;
}

void function_prologue(file, vars)
  FILE *file;
  int vars;
{
  register int regno;
  register int regs = 0;
  register int fp = 0;

  for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
    if (regs_ever_live[regno] && !call_used_regs[regno])
      regs += 4;
  
  if (frame_pointer_needed && vars)
    fp = 4;

  fprintf(file, ";  Allocated %d bytes: ", vars + regs + fp);
  if (vars)
    fprintf(file, "%d bytes of local vars, ", vars);
  if (fp)
    fprintf(file, "frame pointer, ");
  fprintf(file, "%d reg(s)\n", regs/4);
        
  if (fp) 
    {
      fprintf (file, "\th.stor32\t%d(%%r1),%%r2 ", -vars - regs - fp - 4);
      fprintf (file, "\t; frame pointer pushed\n");
      fprintf (file, "\th.mov32 \t%%r2,%%r1\n");
    }

  if (vars + regs + fp)
    fprintf (file, "\th.addi16s\t%r1,%%r1,%d\n", -vars - regs - fp);

  regs = 0;
  for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
    if (regs_ever_live[regno] && !call_used_regs[regno]) 
      {
        fprintf (file, "\th.stor32\t%d(%%r1),%%r%d \t", (regs += 4) - 8 + fp, regno);
        fprintf (file, "; general register pushed on stack\n");
      }
}

void function_epilogue(file, vars)
  FILE *file;
  int vars;
{
  register int regno;
  register int regs;
  register int fp = 0;

  if (frame_pointer_needed && vars)
    fp = 4;
  
  regs = 0;
  for (regno = 0; regno < FIRST_PSEUDO_REGISTER; ++regno)
    if (regs_ever_live[regno] && !call_used_regs[regno]) 
      {
        fprintf (file, "\th.load32u\t%%r%d,%d(%%r1) \t", regno, (regs += 4) - 8 + fp);
        fprintf (file, "; general register poped from stack\n");
      }

  if (fp) 
    {
      fprintf (file, "\th.load32u\t%%r2,-4(%%r1) \t");
      fprintf (file, "; frame pointer poped from stack\n");
    }

  fprintf (file, "\th.jr \t\t%%r11\n");

  if (vars + regs + fp)
    fprintf (file, "\th.addi16s\t%%r1,%%r1,%d\n", vars + regs + fp);
  else
    fprintf (file, "\th.nop\n");

  fprintf (file, ".endproc _%s\n", get_function_name());
}

int log_of_two(int x)
{
  int y;
  
  x >>= 1;
  for (y=0; x; x>>=1, y++)
    ;
      
  return y;
}
