/* Subroutines for insn-output.c for GNU compiler.  OpenRISC 1000 version.
   Copyright (C) 1987, 1992, 1997 Free Software Foundation, Inc
   Contributed by Damjan Lampret <lampret@opencores.org> in 1999.
   Based upon the elxsi (which was based upon the vax port) and sparc port.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include "config.h"
#include <stdio.h>
#include "rtl.h"
#include "tree.h"
#include "flags.h"
#include "regs.h"
#include "hard-reg-set.h"

/*
 * for holding comparison operands.
 */
rtx or1k_cmp_op0, or1k_cmp_op1;
enum machine_mode or1k_cmp_mode;

/*
 * used in function prologue/epilogue generation
 */
extern int leaf_function;

int
print_operand_punct_valid_p(code)
     int code;
{
  switch(code)
    {
      case '(':   /* idea taken from sparc; output nop for %( if
                     not optimizing or the slot is not filled. */
      case '%':
        return 1;
    }
  return 0;
}
                                                  
print_operand_address (file, addr)
     FILE *file;
     register rtx addr;
{
  register rtx reg1, reg2, breg, ireg;
  rtx offset;
 
 retry:
  switch (GET_CODE (addr))
    {

    case MEM:
      if (GET_CODE (XEXP (addr, 0)) == REG)
        fprintf (file, "%s", reg_names[REGNO (addr)]);
      else abort();
      break;

    case REG:
      fprintf (file, "0(%s)", reg_names[REGNO (addr)]);
      break;

    case PLUS:
      reg1 = 0;	reg2 = 0;
      ireg = 0;	breg = 0;
      offset = 0;
      if (GET_CODE (XEXP (addr, 0)) == REG)
	{
	  offset = XEXP (addr, 1);
	  addr = XEXP (addr, 0);
	}
      else if (GET_CODE (XEXP (addr, 1)) == REG)
	{
	  offset = XEXP (addr, 0);
	  addr = XEXP (addr, 1);
	}
      output_address (offset);
      fprintf (file, "(%s)", reg_names[REGNO (addr)]);
      break;

    default:
      /* fprintf(file, "{%d}", GET_CODE (addr)); */
      output_addr_const (file, addr);
    }
}

char *
get_function_name()
{
  char *kind = "function";
  if (current_function_decl != 0
      && TREE_CODE (TREE_TYPE (current_function_decl)) == METHOD_TYPE)
    kind = "method";

  if (current_function_decl == NULL)
    {
      return "top level";
    }
  else
    {
      char *name = (*decl_printable_name) (current_function_decl, &kind);
      return name;
    }
}

/*
 * operands[0] will be the branch target.
 * test will be the comparison type.
 * or1k_cmp_mode is SImode
 * or1k_cmp_op0 and or1k_cmp_op1 are the things to be 
 * compared.
 *
 * This routine has to generate a compare and a branch rtx.
 *
 */
int
gen_cond_branch(operands, test)
     rtx operands[];
     enum rtx_code test;
{
  rtx tmp;

  switch(or1k_cmp_mode)
    {
    case SImode:
      /* generate a sfXX instruction. */

      /* check if operand 1 is an integer that is too big;
         move it into a register if it is. */
/* Not used if sfXXi are available 
*/       if(GET_CODE(or1k_cmp_op1) == CONST_INT)
        {
          int c = INTVAL(or1k_cmp_op1);
          or1k_cmp_op1 = force_reg(SImode, or1k_cmp_op1);
        }
/**/

      /* emit the set instruction. */
      tmp = gen_rtx(SET, VOIDmode, cc0_rtx,
                         gen_rtx(test, SImode,
                                 or1k_cmp_op0, or1k_cmp_op1));
      emit_insn( tmp);
  
      /* emit the branch instruction. */
      tmp = gen_rtx(SET, VOIDmode, 
                              pc_rtx,
                              gen_rtx(IF_THEN_ELSE, VOIDmode,
                                      gen_rtx(NE, VOIDmode, const0_rtx,
                                              cc0_rtx),
                                      gen_rtx(LABEL_REF,VOIDmode,operands[0]), 
                                      pc_rtx));
      emit_jump_insn( tmp);
      break;
    }

}

int
sym_ref_mem_operand(op, mode)
	register rtx op;
	enum machine_mode mode;
{
	if(GET_CODE(op) == MEM) {
		rtx t1 = XEXP(op, 0);
		if (GET_CODE(t1) == SYMBOL_REF)
			return 1;
	}

	return 0;
}

void function_prologue(file, vars)
	FILE *file;
	int vars;
{
	register int regno;
	register int regs = 0;
	register int fp = 0;

	for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
		if (regs_ever_live[regno] && !call_used_regs[regno])
			regs += 4;
	
	if (frame_pointer_needed && vars)
		fp = 4;

	fprintf(file, "#  Allocated %d bytes: ", vars + regs + fp);
	if (vars)
		fprintf(file, "%d bytes of local vars, ", vars);
	if (fp)
		fprintf(file, "frame pointer, ");
	fprintf(file, "%d reg(s)\n", regs/4);
        
        if (fp) {
        	fprintf (file, "\tl.sw     \t%d(r%d),r%d ", -vars - regs - fp - 4, STACK_POINTER_REGNUM, FRAME_POINTER_REGNUM);
        	fprintf (file, "\t# frame pointer pushed\n");
		fprintf (file, "\tl.addi   \tr%d,r%d,0\n", FRAME_POINTER_REGNUM, STACK_POINTER_REGNUM);
	}

	if (vars + regs + fp)
		fprintf (file, "\tl.addi   \tr%d,r%d,%d\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM, -vars - regs - fp);

	regs = 0;
	for (regno = 0; regno < FIRST_PSEUDO_REGISTER; regno++)
		if (regs_ever_live[regno] && !call_used_regs[regno]) {
			fprintf (file, "\tl.sw    \t%d(r%d),r%d \t", (regs += 4) - 8 + fp, STACK_POINTER_REGNUM, regno);
			fprintf (file, "# general register pushed on stack\n");
		}
}

void function_epilogue(file, vars)
	FILE *file;
	int vars;
{
	register int regno;
	register int regs;
	register int fp = 0;

	if (frame_pointer_needed && vars)
		fp = 4;
	
	regs = 0;
	for (regno = 0; regno < FIRST_PSEUDO_REGISTER; ++regno)
		if (regs_ever_live[regno] && !call_used_regs[regno]) {
			fprintf (file, "\tl.lw     \tr%d,%d(r%d) \t", regno, (regs += 4) - 8 + fp, STACK_POINTER_REGNUM);
			fprintf (file, "# general register poped from stack\n");
		}

	if (fp) {
		fprintf (file, "\tl.lw     \tr%d,-4(r%d) \t", FRAME_POINTER_REGNUM, STACK_POINTER_REGNUM);
		fprintf (file, "# frame pointer poped from stack\n");
	}

	fprintf (file, "\tl.jalr  \tr%d\n", LINK_REGNUM);

	if (vars + regs + fp)
		fprintf (file, "\tl.addi   \tr%d,r%d,%d\n", STACK_POINTER_REGNUM, STACK_POINTER_REGNUM, vars + regs + fp);
	else
		fprintf (file, "\tl.nop\n");

	fprintf (file, ".endproc _%s\n", get_function_name());
}

int log_of_two(int x)
{
	int y;
  
	x >>= 1;
	for(y=0; x; x>>=1, y++);
      
	return y;
}

void
print_operand(file, x, code)
    FILE *file;
    rtx x;
    char code;
{ 									
  if (code == 'r' && GET_CODE (x) == MEM && GET_CODE (XEXP (x, 0)) == REG) 
    fprintf (file, "%s", reg_names[REGNO (XEXP (x, 0))]);		
  else if (code == '(')							
    if (dbr_sequence_length())						
      fprintf (file, "\t# delay slot filled");				
    else								
      fprintf (file, "\n\tl.nop\t\t\t# nop delay slot");			
  else if (code == 'C')							
     switch(GET_CODE(x))						
       {								
         case EQ:        fputs("eq", file); break;			
         case NE:        fputs("ne", file); break;			
         case GT:        fputs("gts", file); break;			
         case GE:        fputs("ges", file); break;			
         case LT:        fputs("lts", file); break;			
         case LE:        fputs("les", file); break;			
         case GTU:       fputs("gtu", file); break;			
         case GEU:       fputs("geu", file); break;			
         case LTU:       fputs("ltu", file); break;			
         case LEU:       fputs("leu", file); break;			
         default:							
           abort ();							
       }   								
  else if (GET_CODE (x) == REG)						
    fprintf (file, "%s", reg_names[REGNO (x)]);	
  else if (GET_CODE (x) == MEM)						
	output_address (XEXP (x, 0));					
  else 									
  { 									
	/*debug_rtx(x);*/ 						
	output_addr_const (file, x); 					
   }									
}

