;; Machine description for GNU compiler, OpenRISC 1000 family, OR32 ISA
;; Copyright (C) 1999 Free Software Foundation, Inc.
;; Contributed by Damjan Lampret <lampret@opencores.org> in 1999.
;; Based upon the elxsi (which was based upon the vax port) and sparc port.

;; This file is part of GNU CC.

;; GNU CC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.

;; GNU CC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU CC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

(define_attr "type"
  "unknown,load,store,move,extend,logic,add,mul,shift,compare,branch,jump,fp"
  (const_string "unknown"))

;; Number of instructions
(define_attr "length" "" (const_int 1))

(define_delay (eq_attr "type" "branch,jump")
               [(and (eq_attr "type" "!branch,jump")
		    (eq_attr "length" "1")) (nil) (nil)])

;; (define_function_unit NAME MULTIPLICITY SIMULTANEITY
;;                       TEST READY-DELAY ISSUE-DELAY [CONFLICT-LIST])
;; MULTIPLICITY - Number of functional units of this type
;; SIMULTANEITY - Zero for pipelined functional unit
;; READY-DELAY - Number of cycles before result is available
;; ISSUE-DELAY - Number of cycles before unit can accept new instruction
;;
(define_function_unit "bit_unit" 1 0 (eq_attr "type" "shift") 3 1)
(define_function_unit "lsu" 1 0 (eq_attr "type" "load") 3 3)
(define_function_unit "lsu" 1 0 (eq_attr "type" "store") 2 1)
(define_function_unit "alu" 1 0 (eq_attr "type" "add,logic,extend,move,compare") 2 1)
(define_function_unit "mul_unit" 1 0 (eq_attr "type" "mul") 16 16)

;;
;; Experimental FP stuff
;;

(define_insn "addsf3"
  [(set (match_operand:SF 0 "register_operand" "=r")
        (plus:SF (match_operand:SF 1 "register_operand" "r")
                 (match_operand:SF 2 "register_operand" "r")))]
  ""
  "l.fadd.s\\t%0,%1,%2"
  [(set_attr "type"     "fp")
   (set_attr "length"   "1")])

(define_insn "subsf3"
  [(set (match_operand:SF 0 "register_operand" "=r")
        (minus:SF (match_operand:SF 1 "register_operand" "r")
                 (match_operand:SF 2 "register_operand" "r")))]
  ""
  "l.fsub.s\\t%0,%1,%2"
  [(set_attr "type"     "fp")
   (set_attr "length"   "1")])

(define_insn "mulsf3"
  [(set (match_operand:SF 0 "register_operand" "=r")
        (mult:SF (match_operand:SF 1 "register_operand" "r")
                 (match_operand:SF 2 "register_operand" "r")))]
  ""
  "l.fmul.s\\t%0,%1,%2"
  [(set_attr "type"     "fp")
   (set_attr "length"   "1")])

;;
;; Compares
;;

(define_expand "cmpsi"
  [(set (cc0)
        (compare:CC (match_operand:SI 0 "register_operand"  "r")
                    (match_operand:SI 1 "nonmemory_operand" "rI")))]
  ""
  "
{
      or1k_cmp_op0 = operands[0];
      or1k_cmp_op1 = operands[1];
      or1k_cmp_mode = SImode;
      DONE;
}")

(define_expand "tstsi"
  [(set (cc0)
        (match_operand:SI 0 "register_operand" "r"))]
  ""
  "
{
      or1k_cmp_op0 = operands[0];
      or1k_cmp_op1 = const0_rtx;
      or1k_cmp_mode = SImode;
      DONE;
}")

;;
;; Moves
;;

(define_insn "movdf"  [(set (match_operand:DF 0 "general_operand" "=r,r,m")
        (match_operand:DF 1 "general_operand" "r,m,r"))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.addi  \\t%0,%1,0\\t # fake insn for handling DF\";
     case 1:
       return \"l.lw    \\t%0,%1\\t # fake insn for handling DF\";
     case 2:
       return \"l.sw    \\t%0,%1\\t # fake insn for handling DF\";
}"
[(set_attr "type" "move,load,store")
   (set_attr "length" "1,1,1")])

(define_insn "movsf"
  [(set (match_operand:SF 0 "general_operand" "=r,r,m")
        (match_operand:SF 1 "general_operand"  "r,m,r"))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.addi  \\t%0,%1,0\\t # move reg to reg\";
     case 1:
       return \"l.lw    \\t%0,%1\";
     case 2:
       return \"l.sw    \\t%0,%1\";
}"
[(set_attr "type" "move,load,store")
   (set_attr "length" "1,1,1")])

(define_insn "movsi"
  [(set (match_operand:SI 0 "nonimmediate_operand" "=r,r,m,r,r,r")
	(match_operand:SI 1 "general_operand"       "r,m,r,I,K,i"))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.addi  \\t%0,%1,0\\t # move reg to reg\";
     case 1:
       return \"l.lw    \\t%0,%1\";
     case 2:
       return \"l.sw    \\t%0,%1\";
     case 3:
       return \"l.addi  \\t%0,r0,%1\\t # move immediate\";
     case 4:
       return \"l.ori   \\t%0,r0,%1\\t # move immediate\";
     case 5:
       return \"l.movhi \\t%0,hi(%1)\\t # move immediate (high)\;l.ori  \\t%0,%0,lo(%1)\\t # move immediate (low)\";
   }
  "
  [(set_attr "type" "add,load,store,add,logic,move")
   (set_attr "length" "1,1,1,1,1,2")])

(define_insn "movhi"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=r,m,r,r,r")
	(match_operand:HI 1 "general_operand"       "m,r,r,I,K"))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.lhz   \\t%0,%1\";
     case 1:
       return \"l.sh    \\t%0,%1\";
     case 2:
       return \"l.addi  \\t%0,%1,0\\t # move reg to reg\";
     case 3:
       return \"l.addi  \\t%0,r0,%1\\t # move immediate\";
     case 4:
       return \"l.ori   \\t%0,r0,%1\\t # move immediate\";
   }
  "
  [(set_attr "type" "load,store,add,add,logic")])


(define_insn "movqi"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=r,m,r,r,r")
	(match_operand:QI 1 "general_operand"       "m,r,r,I,K"))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.lbz   \\t%0,%1\";
     case 1:
       return \"l.sb    \\t%0,%1\";
     case 2:
       return \"l.addi  \\t%0,%1,0\\t # move reg to reg\";
     case 3:
       return \"l.addi  \\t%0,r0,%1\\t # move immediate\";
     case 4:
       return \"l.ori   \\t%0,r0,%1\\t # move immediate\";
   }
  "
  [(set_attr "type" "load,store,add,add,logic")])

;;
;; Extension and truncation insns.
;;

;;(define_insn "truncsiqi2"
;;  [(set (match_operand:QI 0 "nonimmediate_operand"      "=r,r,m")
;;	(truncate:QI (match_operand:SI 1 "general_operand" "r,m,r")))]
;;  ""
;;  "*
;;   switch (which_alternative) {
;;     case 0:
;;	return \"l.andi  \\t%0,%1,0x00ff\\t # truncsiqi2\";
;;     case 1:
;;	return \"l.lbz   \\t%0,%1\";
;;     case 2:
;;	return \"l.sh    \\t%0,%1\";
;;   }
;;  "
;;  [(set_attr "type" "logic,load,store")
;;   (set_attr "length" "1,1,1")])

;;(define_insn "truncsihi2"
;;  [(set (match_operand:HI 0 "nonimmediate_operand"      "=r,m")
;;	(truncate:HI (match_operand:SI 1 "general_operand" "m,r")))]
;;  ""
;;  "*
;;   switch (which_alternative) {
;;     case 0:
;;        return \"l.lhz   \\t%0,%1\";
;;     case 1:
;;        return \"l.sh    \\t%0,%1\";
;;   }
;;  "
;;  [(set_attr "type" "load,store")])

;;(define_insn "trunchiqi2"
;;  [(set (match_operand:QI 0 "nonimmediate_operand"      "=r,r,m")
;;	(truncate:QI (match_operand:HI 1 "general_operand" "r,m,r")))]
;;  ""
;;  "*
;;   switch (which_alternative) {
;;     case 0:
;;        return \"l.andi  \\t%0,%1,0x00ff\\t # trunchiqi2\";
;;     case 1:
;;        return \"l.lbz   \\t%0,%1\";
;;     case 2:
;;        return \"l.sb    \\t%0,%1\";
;;   }
;;  "
;;  [(set_attr "type" "logic,load,store")])

(define_insn "extendhisi2"
  [(set (match_operand:SI 0 "register_operand"               "=r,r")
	(sign_extend:SI (match_operand:HI 1 "general_operand" "m,r")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.lhs   \\t%0,%1\\t # extendhisi2\";
     case 1:
       return \"l.slli  \\t%0,%1,16\;l.srai  \\t%0,%0,16\\t # extendhisi2\";
   }
  "
  [(set_attr "type" "load,shift")
   (set_attr "length" "1,2")])

(define_insn "extendqisi2"
  [(set (match_operand:SI 0 "register_operand"               "=r,r")
	(sign_extend:SI (match_operand:QI 1 "general_operand" "m,r")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.lbs   \\t%0,%1\\t # extendqisi2\";
     case 1:
       return \"l.slli  \\t%0,%1,24\;l.srai  \\t%0,%0,24\\t # extendqisi2\";
   }
  "
  [(set_attr "type" "load,shift")
   (set_attr "length" "1,2")])

(define_insn "zero_extendhisi2"
  [(set (match_operand:SI 0 "register_operand"               "=r,r")
	(zero_extend:SI (match_operand:HI 1 "general_operand" "m,r")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.lhz   \\t%0,%1\\t # zero_extendhisi2\";
     case 1:
       return \"l.andi  \\t%0,%1,0xffff\\t # zero_extendhisi2\";
   }
  "
  [(set_attr "type" "load,extend")])

(define_insn "zero_extendqisi2"
  [(set (match_operand:SI 0 "register_operand"               "=r,r")
	(zero_extend:SI (match_operand:QI 1 "general_operand" "m,r")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.lbz   \\t%0,%1\\t # zero_extendqisi2\";
     case 1:
       return \"l.andi  \\t%0,%1,0x00ff\\t # zero_extendqisi2\";
   }
  "
  [(set_attr "type" "load,logic")])

;;
;; Shift operations
;;

(define_insn "ashlsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
        (ashift:SI (match_operand:SI 1 "register_operand" "r,r")
                   (match_operand:SI 2 "nonmemory_operand" "r,L")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.sll   \\t%0,%1,%2\";
     case 1:
       return \"l.slli  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "shift,shift")])

(define_insn "ashrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
        (ashiftrt:SI (match_operand:SI 1 "register_operand" "r,r")
                   (match_operand:SI 2 "nonmemory_operand" "r,L")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.sra   \\t%0,%1,%2\";
     case 1:
       return \"l.srai  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "shift,shift")])

(define_insn "lshrsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
        (lshiftrt:SI (match_operand:SI 1 "register_operand" "r,r")
                   (match_operand:SI 2 "nonmemory_operand" "r,L")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.srl   \\t%0,%1,%2\";
     case 1:
       return \"l.srli  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "shift,shift")])

;;
;; Logical bitwise operations
;;

(define_insn "andsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(and:SI (match_operand:SI 1 "register_operand" "%r,r")
		(match_operand:SI 2 "nonmemory_operand" "r,K")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.and   \\t%0,%1,%2\";
     case 1:
       return \"l.andi  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "logic,logic")])

(define_insn "iorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(ior:SI (match_operand:SI 1 "register_operand" "%r,r")
		(match_operand:SI 2 "nonmemory_operand" "r,K")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.or    \\t%0,%1,%2\";
     case 1:
       return \"l.ori   \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "logic,logic")])

(define_insn "xorsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(xor:SI (match_operand:SI 1 "register_operand" "%r,r")
		(match_operand:SI 2 "nonmemory_operand" "r,I")))]
  ""
  "*
   switch(which_alternative) {
     case 0:
       return \"l.xor   \\t%0,%1,%2\";
     case 1:
       return \"l.xori  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "logic,logic")])

(define_insn "one_cmplqi2"
  [(set (match_operand:QI 0 "register_operand" "=r")
	(not:QI (match_operand:QI 1 "register_operand" "r")))]
  ""
  "l.xori  \\t%0,%1,0x00ff"
  [(set_attr "type" "logic")])

;; gen_rtx_XOR(Himode ... mising XOR for HImode define_insn
;;(define_expand "one_cmplhi2"
;;  [(set (match_operand:HI 0 "register_operand" "=r")
;;	(not:HI (match_operand:HI 1 "register_operand" "r")))]
;;  ""
;;  "
;;   {
;;        if ((reload_in_progress | reload_completed) == 0) {
;;               rtx tmpreg, tmp;
;;
;;                tmp = gen_rtx(SET, HImode, tmpreg = gen_reg_rtx(HImode),
;;                                                GEN_INT(0xffff));
;;                emit_insn(tmp);
;;               emit_insn(gen_rtx(SET, HImode, operands[0],
;;                                        gen_rtx_XOR(HImode, operands[1], tmpreg)));
;;                DONE;
;;        }
;;   }
;;  ")

(define_insn "one_cmplsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(not:SI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "l.xori  \\t%0,%1,-1"
  [(set_attr "type" "logic")])

;;
;; Arithmetic operations 
;;

(define_insn "negsi2"
  [(set (match_operand:SI 0 "register_operand" "=r")
	(neg:SI (match_operand:SI 1 "register_operand" "r")))]
  ""
  "*
    return \"l.sub   \\t%0,r0,%1\";
  "
  [(set_attr "type" "add")])

(define_insn "addsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(plus:SI (match_operand:SI 1 "register_operand" "%r,r")
		 (match_operand:SI 2 "nonmemory_operand" "r,I")))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.add   \\t%0,%1,%2\";
     case 1:
       return \"l.addi  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "add,add")])

(define_insn "subsi3"
  [(set (match_operand:SI 0 "register_operand" "=r,r")
	(minus:SI (match_operand:SI 1 "register_operand" "r,r")
		  (match_operand:SI 2 "nonmemory_operand" "r,I")))]
  ""
  "*
  switch(which_alternative)
   {
     case 0:
       return \"l.sub   \\t%0,%1,%2\";
     case 1:
       return \"l.subi  \\t%0,%1,%2\";
   }
  "
  [(set_attr "type" "add,add")])

;; mul and div

(define_insn "mulsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (mult:SI (match_operand:SI 1 "register_operand" "r")
                 (match_operand:SI 2 "register_operand" "r")))]
  ""
  "l.mul   \\t%0,%1,%2"
  [(set_attr "type" "mul")])

(define_insn "divsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (div:SI (match_operand:SI 1 "register_operand" "r")
                 (match_operand:SI 2 "register_operand" "r")))]
  ""
  "l.div   \\t%0,%1,%2"
  [(set_attr "type" "mul")])

(define_insn "udivsi3"
  [(set (match_operand:SI 0 "register_operand" "=r")
        (udiv:SI (match_operand:SI 1 "register_operand" "r")
                 (match_operand:SI 2 "register_operand" "r")))]
  ""
  "l.divu  \\t%0,%1,%2"
  [(set_attr "type" "mul")])

;; jumps 

(define_insn "jump"
  [(set (pc)
	(label_ref (match_operand 0 "" "")))]
  ""
  "l.j     \\t%l0%("
  [(set_attr "type" "jump")])

(define_insn "indirect_jump"
  [(set (pc) (match_operand:SI 0 "register_operand" "r"))]
  ""
  "l.jr    \\t%0%("
  [(set_attr "type" "jump")])

;;
;; Conditional Branches
;; 

(define_expand "beq"
  [(set (pc) (if_then_else (eq (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, EQ); DONE; }")

(define_expand "bne"
  [(set (pc) (if_then_else (ne (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, NE); DONE; }")

(define_expand "blt"
  [(set (pc) (if_then_else (lt (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, LT); DONE; }")

(define_expand "bgt"
  [(set (pc) (if_then_else (gt (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, GT); DONE; }")

(define_expand "ble"
  [(set (pc) (if_then_else (le (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, LE); DONE; }")

(define_expand "bge"
  [(set (pc) (if_then_else (ge (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, GE); DONE; }")

(define_expand "bltu"
  [(set (pc) (if_then_else (ltu (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, LTU); DONE; }")

(define_expand "bgtu"
  [(set (pc) (if_then_else (gtu (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, GTU); DONE; }")

(define_expand "bleu"
  [(set (pc) (if_then_else (leu (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, LEU); DONE; }")

(define_expand "bgeu"
  [(set (pc) (if_then_else (geu (cc0) (const_int 0))
     (label_ref (match_operand 0 "" "")) (pc)))]
  "" "{ gen_cond_branch(operands, GEU); DONE; }")

;;
;; Output insns for conditional branches; RTL for these is 
;; generated in gen_cond_branch().
;;

(define_insn "branch_output"
  [(set (pc) (if_then_else (eq (const_int 0)
			    (cc0))
                           (label_ref (match_operand 0 "" "")) (pc)))]
  ""
  "l.bnf   \\t%0%("
  [(set_attr "type" "branch")])

(define_insn "branch_output_rev"
  [(set (pc) (if_then_else (ne (const_int 0)
                            (cc0))
                           (label_ref (match_operand 0 "" "")) (pc)))]
  ""
  "l.bf    \\t%0%("
  [(set_attr "type" "branch")])

;;
;; Output insns for set intructions; RTL for these is
;; generated in gen_cond_branch().
;;

(define_insn "set_output1"
  [(set (cc0)
        (match_operator:SI 0 "comparison_operator"
         [(match_operand:SI 1 "register_operand" "r")
          (match_operand:SI 2 "register_operand" "r")]))]
  ""
  "l.sf%C0  \\t%1,%2"
  [(set_attr "type" "compare")])

;;(define_insn "set_output2"
;;  [(set (cc0)
;;        (match_operator:SI 0 "comparison_operator"
;;         [(match_operand:SI 1 "register_operand" "r")
;;          (match_operand:SI 2 "immediate_operand" "I")]))]
;;  ""
;;  "l.sf%C0i  \\t%1,%2"
;;  [(set_attr "type" "compare")])

;;
;; Calls
;;
(define_insn "call_XXX"
  [(call (mem:SI (match_operand 0 "sym_ref_mem_operand" ""))
         (match_operand 1 "" "i"))]
  ""
  "XXX\\t%S0%("
  [(set_attr "type" "jump")])

(define_insn "call_YYY"
  [(parallel [(call (mem:SI (match_operand 0 "sym_ref_mem_operand" ""))
                    (match_operand 1 "" "i"))
	      (clobber (reg:SI 11))])]
  ""
  "l.jal   \\t%S0%("
  [(set_attr "type" "jump")])

 
(define_insn "call_value"
[(parallel [(set (match_operand 0 "register_operand" "=r")
                  (call (match_operand 1 "sym_ref_mem_operand" "")
                        (match_operand 2 "" "i")))
            (clobber (reg:SI 11))])]
  ""
  "l.jal   \\t%S1%("
  [(set_attr "type" "jump")])

(define_insn "call"
[(parallel [(call (match_operand 0 "sym_ref_mem_operand" "")
                  (match_operand 1 "" "i"))
            (clobber (reg:SI 11))])]
  ""
  "l.jal   \\t%S0%("
  [(set_attr "type" "jump")])

(define_insn "call_value_indirect"
  [(parallel [(set (match_operand 0 "register_operand" "=r")
                   (call (mem:SI (match_operand 1 "register_operand" "r"))
                         (match_operand 2 "" "i")))
              (clobber (reg:SI 11))])]
  ""
  "l.jalr  \\t%1%("
  [(set_attr "type" "jump")])

(define_insn "call_indirect"
[(parallel [(call (mem:SI (match_operand 0 "register_operand" "r"))
                  (match_operand 1 "" "i"))
              (clobber (reg:SI 11))])]
  ""
  "l.jalr  \\t%0%("
  [(set_attr "type" "jump")])

(define_insn "tablejump"
  [(set (pc) (match_operand:SI 0 "register_operand" "r"))
   (use (label_ref (match_operand 1 "" "")))]
  ""
  "l.jalr  \\t%0%("
  [(set_attr "type" "jump")])

(define_insn "nop"
  [(const_int 0)]
  ""
  "l.nop"
  [(set_attr "type" "logic")])
